@@ -11,7 +11,7 @@ from typing import Dict, List, Tuple
 import numpy as np
 import quaternion
 import yaml
-from arcor2_calibration_data import CALIBRATION_URL, SERVICE_NAME, Corner, MarkerCorners
+from arcor2_calibration_data import CALIBRATION_URL, SERVICE_NAME, Corner, EstimatedPose, MarkerCorners
 from arcor2_calibration_data.client import CalibrateRobotArgs
 from dataclasses_jsonschema import ValidationError
 from flask import jsonify, request
@@ -24,6 +24,7 @@ from arcor2.flask import RespT, create_app, run_app
 from arcor2.helpers import port_from_url
 from arcor2.logging import get_logger
 from arcor2.urdf import urdf_from_url
+from arcor2_calibration import calibration
 from arcor2_calibration.calibration import detect_corners, estimate_camera_pose
 from arcor2_calibration.quaternions import weighted_average_quaternions
 from arcor2_calibration.robot import calibrate_robot
@@ -34,6 +35,12 @@ logger.propagate = False
 MARKERS: Dict[int, Pose] = {}
 MARKER_SIZE = 0.1
 
+MIN_DIST = 0.3
+MAX_DIST = 2.0
+
+MIN_THETA = 2.0
+MAX_THETA = math.pi
+
 _mock: bool = False
 
 app = create_app(__name__)
@@ -52,6 +59,14 @@ def dist_matrix_from_request() -> List[float]:
     return [float(val) for val in request.args.getlist("distCoefs")]
 
 
+def normalize(val: float, min_val: float, max_val: float) -> float:
+
+    assert max_val > min_val
+    res = (min(max(abs(val), min_val), max_val) - min_val) / (max_val - min_val)
+    assert 0 <= res <= 1
+    return res
+
+
 @app.route("/calibrate/robot", methods=["PUT"])
 def put_calibrate_robot() -> RespT:
     """Get calibration (camera pose wrt. marker)
@@ -86,7 +101,7 @@ def put_calibrate_robot() -> RespT:
     """
 
     image = Image.open(request.files["image"].stream)
-    args = CalibrateRobotArgs.from_json(request.files["args"].read())
+    args = CalibrateRobotArgs.from_json(request.files["args"].stream.read().decode())
 
     if _mock:
         time.sleep(5)
@@ -265,7 +280,13 @@ def get_calibration() -> RespT:
               content:
                 application/json:
                   schema:
-                    $ref: Pose
+                    $ref: EstimatedPose
+            404:
+              description: No marker found.
+              content:
+                application/json:
+                  schema:
+                    type: string
 
     """
 
@@ -277,6 +298,7 @@ def get_calibration() -> RespT:
 
     if _mock:
         time.sleep(0.5)
+        quality = random.uniform(0, 1)
         pose = Pose(Position(random.uniform(-0.5, 0.5), random.uniform(-0.5, 0.5), random.uniform(0.2, 1)))
     else:
         poses = estimate_camera_pose(camera_matrix, dist_matrix, image, MARKER_SIZE)
@@ -284,6 +306,7 @@ def get_calibration() -> RespT:
         if not poses:
             return jsonify("No marker detected."), 404
 
+        quality_dict: Dict[int, float] = {k: 0.0 for k, v in MARKERS.items()}
         known_markers: List[Tuple[Pose, float]] = []
 
         # apply configured marker offset from origin to the detected poses
@@ -295,12 +318,26 @@ def get_calibration() -> RespT:
                 continue
 
             mpose = poses[marker_id]
-            weight = 1.0 / math.sqrt(mpose.position.x ** 2 + mpose.position.y ** 2 + mpose.position.z ** 2)
-            known_markers.append((tr.make_pose_abs(cpose, mpose), weight))
+            dist = math.sqrt(mpose.position.x ** 2 + mpose.position.y ** 2 + mpose.position.z ** 2)
+
+            # the closer the theta is to pi, the higher quality we get
+            theta = quaternion.as_spherical_coords(mpose.orientation.as_quaternion())[0]
+            ori_marker_quality = normalize(abs(theta), MIN_THETA, MAX_THETA)
+
+            # the closer the marker is, the higher quality we get
+            dist_marker_quality = 1.0 - normalize(dist, MIN_DIST, MAX_DIST)
+
+            marker_quality = (ori_marker_quality + dist_marker_quality) / 2
+
+            quality_dict[marker_id] = marker_quality
+            known_markers.append((tr.make_pose_abs(cpose, mpose), marker_quality))
+
             logger.debug(f"Known marker       : {marker_id}")
             logger.debug(f"...original pose   : {poses[marker_id]}")
             logger.debug(f"...transformed pose: {poses[marker_id]}")
-            logger.debug(f"...weight          : {weight}")
+            logger.debug(f"...dist quality    : {dist_marker_quality:.3f}")
+            logger.debug(f"...ori quality     : {ori_marker_quality:.3f}")
+            logger.debug(f"...overall quality : {marker_quality:.3f}")
 
         if not known_markers:
             return jsonify("No known marker detected."), 404
@@ -308,7 +345,6 @@ def get_calibration() -> RespT:
         weights = [marker[1] for marker in known_markers]
 
         # combine all detections
-        # TODO this is just initial (naive) solution with weight equal to distance to the origin
         pose = Pose()
         for mpose, weight in known_markers:
             pose.position += mpose.position * weight
@@ -319,20 +355,21 @@ def get_calibration() -> RespT:
             quaternion.from_float_array(weighted_average_quaternions(quaternions, np.array(weights)))
         )
 
+        quality = np.mean(list(quality_dict.values()))
+
     inverse = request.args.get("inverse", default="false") == "true"
 
     if inverse:
         logger.debug("Inverting the output pose.")
         pose = pose.inversed()
 
-    return jsonify(pose), 200
+    return jsonify(EstimatedPose(pose, quality)), 200
 
 
 def main() -> None:
 
     parser = argparse.ArgumentParser(description=SERVICE_NAME)
-    parser.add_argument("-s", "--swagger", action="store_true", default=False)
-    parser.add_argument("-m", "--mock", action="store_true", default=False)
+
     parser.add_argument(
         "-d",
         "--debug",
@@ -342,15 +379,18 @@ def main() -> None:
         default=logging.INFO,
     )
 
-    group = parser.add_mutually_exclusive_group()
+    group = parser.add_mutually_exclusive_group(required=True)
+
     group.add_argument(
         "--config-file",
         "-c",
         type=argparse.FileType("r"),
-        default=sys.stdin,
         help="Config file name containing a valid YAML configuration.",
     )
-    group.add_argument("yaml-config", nargs="?", type=str, help="Input string containing a valid YAML configuration.")
+
+    sub_group = group.add_mutually_exclusive_group()
+    sub_group.add_argument("-s", "--swagger", action="store_true", default=False)
+    sub_group.add_argument("-m", "--mock", action="store_true", default=False)
 
     args = parser.parse_args()
 
@@ -358,18 +398,20 @@ def main() -> None:
 
     if not (args.swagger or args.mock):
 
-        try:
-            data = args.config_file.read()
-        except AttributeError:
-            data = args.yaml_config
+        data = args.config_file.read()
 
         global MARKER_SIZE
+        global MIN_DIST
+        global MAX_DIST
 
         try:
 
             config = yaml.safe_load(data)
 
-            MARKER_SIZE = float(config["marker_size"])
+            MARKER_SIZE = float(config.get("marker_size", MARKER_SIZE))
+            MIN_DIST = float(config.get("min_dist", MIN_DIST))
+            MAX_DIST = float(config.get("max_dist", MAX_DIST))
+            calibration.BLUR_THRESHOLD = float(config.get("blur_threshold", calibration.BLUR_THRESHOLD))
 
             for marker_id, marker in config["markers"].items():
                 MARKERS[int(marker_id)] = Pose.from_dict(marker["pose"])
@@ -382,6 +424,10 @@ def main() -> None:
             logger.exception("Failed to load the configuration file.")
             sys.exit(1)
 
+    if not (MAX_DIST > MIN_DIST):
+        logger.error("'max_dist' have to be bigger than 'min_dist'.")
+        sys.exit(1)
+
     global _mock
     _mock = args.mock
     if _mock:
@@ -393,7 +439,7 @@ def main() -> None:
         arcor2_calibration.version(),
         arcor2_calibration.version(),
         port_from_url(CALIBRATION_URL),
-        [Pose, CalibrateRobotArgs, MarkerCorners],
+        [Pose, CalibrateRobotArgs, MarkerCorners, EstimatedPose],
         args.swagger,
     )
 
